'use strict';

var fs = require('fs');
var path = require('path');
var chalk = require('chalk');
var yargs = require('yargs');
var remark = require('remark');
var toc = require('remark-toc');
var reactPlugin = require('@structured-types/react-plugin');
var propTypesPlugin = require('@structured-types/prop-types-plugin');
var api = require('@structured-types/api');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () {
            return e[k];
          }
        });
      }
    });
  }
  n['default'] = e;
  return Object.freeze(n);
}

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var yargs__namespace = /*#__PURE__*/_interopNamespace(yargs);
var remark__default = /*#__PURE__*/_interopDefaultLegacy(remark);
var toc__default = /*#__PURE__*/_interopDefaultLegacy(toc);
var reactPlugin__default = /*#__PURE__*/_interopDefaultLegacy(reactPlugin);
var propTypesPlugin__default = /*#__PURE__*/_interopDefaultLegacy(propTypesPlugin);

const trimQuotes = (txt) => txt ? txt.replace(/['"]+/g, '') : txt;
const extractCustomTag = (node, tagName) => {
    const nodes = node.children &&
        node.children.filter((child) => child.type === 'html' &&
            child.value &&
            child.value.startsWith(`<${tagName}`));
    return nodes
        ? nodes
            .map((section) => ({
            section,
            match: section.value
                ? section.value.match(/(\w+)\s*=\s*((["'])(.*?)\3|([^>\s]*)(?=\s|\/>))(?=[^<]*>)/g)
                : undefined,
        }))
            .map(({ section, match }) => ({
            attrs: { section, tagName, node },
            attributes: match
                ? match.map((m) => m.split('=').map((value) => trimQuotes(value)))
                : undefined,
        }))
        : undefined;
};
const inlineNewContent = ({ section, tagName, node }, newContent) => {
    const startTag = `<!-- START-${tagName.toUpperCase()} -->`;
    const endTag = `<!-- END-${tagName.toUpperCase()} -->`;
    const index = node.children ? node.children.indexOf(section) : -1;
    if (index === -1) {
        return;
    }
    let deleteNodes = 0;
    let foundComment = false;
    if (node.children) {
        for (let i = index + 1; i < node.children.length; i += 1) {
            const childNode = node.children[i];
            if (childNode.type == 'html' && childNode.value == startTag) {
                foundComment = true;
            }
            if (foundComment) {
                deleteNodes += 1;
            }
            if (childNode &&
                childNode.type == 'html' &&
                (childNode.value == endTag ||
                    (childNode.value &&
                        childNode.value.toLowerCase().startsWith(`<${tagName}`)))) {
                break;
            }
        }
        node.children.splice(index + 1, deleteNodes, ...[
            {
                type: 'html',
                value: startTag,
            },
            ...newContent,
            {
                type: 'html',
                value: endTag,
            },
        ]);
    }
};

const insertOverview = () => {
    return (node) => {
        const sections = extractCustomTag(node, 'package-section');
        if (sections) {
            sections.forEach(({ attrs, attributes }) => {
                if (attributes) {
                    const file = attributes.find((attribute) => attribute[0] === 'file');
                    const sectionAttr = attributes.find((attribute) => attribute[0] === 'section');
                    const sectionName = sectionAttr
                        ? sectionAttr[1].toLowerCase()
                        : 'overview';
                    if (file) {
                        const content = fs__default['default'].readFileSync(file[1], 'utf8');
                        const overviewNodes = [];
                        const packageJSONFilename = path__default['default'].resolve(path__default['default'].join(path__default['default'].dirname(file[1]), 'package.json'));
                        if (fs__default['default'].existsSync(packageJSONFilename)) {
                            const packageJSON = JSON.parse(fs__default['default'].readFileSync(packageJSONFilename, 'utf8'));
                            const { name, description, repository: { directory }, } = packageJSON;
                            const rootRepoDir = `https://github.com/ccontrols/component-controls/blob/master/${directory}`;
                            overviewNodes.push({
                                type: 'heading',
                                depth: 2,
                                children: [
                                    {
                                        type: 'link',
                                        url: rootRepoDir,
                                        children: [{ type: 'text', value: name }],
                                    },
                                ],
                            });
                            overviewNodes.push({
                                type: 'paragraph',
                                children: [{ type: 'text', value: description }],
                            });
                        }
                        remark__default['default']()
                            .use(() => (sectionNode) => {
                            let foundDepth = 0;
                            if (sectionNode.children) {
                                for (const section of sectionNode.children) {
                                    if (section.type === 'heading') {
                                        if (section.children &&
                                            section.children.length > 0 &&
                                            section.children[0].type === 'text' &&
                                            section.children[0].value &&
                                            section.children[0].value.toLowerCase() === sectionName) {
                                            foundDepth = section.depth;
                                        }
                                        else if (foundDepth >= section.depth) {
                                            break;
                                        }
                                        else if (foundDepth > 0) {
                                            overviewNodes.push(section);
                                        }
                                    }
                                    else if (foundDepth > 0) {
                                        overviewNodes.push(section);
                                    }
                                }
                            }
                        })
                            .process(content);
                        inlineNewContent(attrs, overviewNodes);
                    }
                }
            });
        }
    };
};

const createPropsRow = ({ name, isOptional, type, value, description }, hasValues) => {
    const children = [
        {
            type: 'tableCell',
            children: [
                { type: 'inlineCode', value: `${name}${isOptional ? '' : '*'}` },
            ],
        },
        {
            type: 'tableCell',
            children: type,
        },
    ];
    if (hasValues) {
        children.push({
            type: 'tableCell',
            children: [
                {
                    type: 'text',
                    value: typeof value !== 'undefined' ? value.toString() : '',
                },
            ],
        });
    }
    children.push({
        type: 'tableCell',
        children: [{ type: 'text', value: description || '' }],
    });
    return {
        type: 'tableRow',
        children,
    };
};
const createPropsTable = (title, children) => {
    const propsTable = [];
    let table = undefined;
    let hasValues = false;
    if (children) {
        propsTable.push({
            type: 'paragraph',
            children: [
                {
                    type: 'heading',
                    depth: 3,
                    children: [
                        {
                            type: 'text',
                            value: title,
                        },
                    ],
                },
            ],
        });
        const columns = [
            { type: 'tableCell', children: [{ type: 'text', value: 'Name' }] },
            { type: 'tableCell', children: [{ type: 'text', value: 'Type' }] },
        ];
        hasValues = children.some((item) => item.value !== undefined);
        if (hasValues) {
            columns.push({
                type: 'tableCell',
                children: [{ type: 'text', value: 'Value' }],
            });
        }
        columns.push({
            type: 'tableCell',
            children: [{ type: 'text', value: 'Description' }],
        });
        table = {
            type: 'table',
            children: [
                {
                    type: 'tableRow',
                    children: columns,
                },
            ],
        };
        propsTable.push({
            type: 'paragraph',
            children: [table],
        });
        // eslint-disable-next-line prefer-spread
        table.children.push(...children.map((child) => createPropsRow(child, hasValues)));
    }
    return { propsTable, table, hasValues };
};

/* eslint-disable prefer-spread */
class ExtractProps {
    constructor(files, names) {
        this.topLevelProps = {};
        this.files = files;
        this.names = names;
    }
    extractPropTable(props, title) {
        const items = props.map((prop) => ({
            name: prop.name || '',
            isOptional: prop.optional,
            type: api.isFunctionProp(prop)
                ? api.isFunctionProp(prop)
                : this.extractPropType(prop),
            description: prop.description || '',
            value: api.hasValue(prop) ? prop.value : undefined,
        }));
        return createPropsTable(title, items);
    }
    extractFunction(node, _extractTable = true) {
        const result = [];
        if (node.name) {
            const declaration = {
                type: 'paragraph',
                children: [],
            };
            result.push(declaration);
            declaration.children.push({
                type: 'strong',
                children: [
                    {
                        type: 'text',
                        value: 'function',
                    },
                ],
            });
            declaration.children.push({
                type: 'text',
                value: ' ',
            });
            declaration.children.push({
                type: 'text',
                value: node.name,
            });
            declaration.children.push({
                type: 'text',
                value: '(',
            });
            if (node.parameters) {
                for (let i = 0; i < node.parameters.length; i += 1) {
                    const p = node.parameters[i];
                    if (i > 0) {
                        declaration.children.push({
                            type: 'text',
                            value: ', ',
                        });
                    }
                    declaration.children.push({
                        type: 'inlineCode',
                        value: p.name,
                    });
                    if (!p.optional) {
                        declaration.children.push({
                            type: 'text',
                            value: '*',
                        });
                    }
                    declaration.children.push({
                        type: 'text',
                        value: ': ',
                    });
                    declaration.children.push(...this.extractPropType(p));
                }
            }
            declaration.children.push({
                type: 'text',
                value: ')',
            });
            if (node.returns) {
                declaration.children.push({
                    type: 'text',
                    value: ': ',
                });
                declaration.children.push(...this.extractPropType(node.returns));
            }
            declaration.children.push({
                type: 'text',
                value: ';',
            });
            if (node.parameters) {
                const { propsTable, table, hasValues } = this.extractPropTable(node.parameters, 'parameters');
                if (table && node.returns && node.returns.kind !== api.PropKind.Void) {
                    table.children.push(createPropsRow({
                        name: 'returns',
                        isOptional: true,
                        type: this.extractPropType(node.returns),
                        description: node.returns.description || '',
                    }, hasValues));
                }
                result.push(...propsTable);
            }
        }
        return result;
    }
    extractInterface(prop) {
        var _a;
        const result = [];
        if (prop.name) {
            const declaration = {
                type: 'paragraph',
                children: [],
            };
            result.push(declaration);
            if ((_a = prop.extends) === null || _a === void 0 ? void 0 : _a.length) {
                declaration.children.push({
                    type: 'strong',
                    children: [
                        {
                            type: 'text',
                            value: ' extends ',
                        },
                        ...prop.extends.reduce((acc, p) => [
                            ...acc,
                            ...this.extractPropType(p),
                        ], []),
                    ],
                });
            }
            if (api.hasProperties(prop) && prop.properties) {
                const { propsTable } = this.extractPropTable(prop.properties, 'properties');
                result.push(...propsTable);
            }
        }
        return result;
    }
    propLink(type) {
        const typeText = [
            {
                type: 'text',
                value: type,
            },
        ];
        if (typeof type === 'string') {
            const propName = type.startsWith(':') ? type.substring(1) : type;
            if (this.topLevelProps[propName]) {
                return [
                    {
                        type: 'link',
                        url: `#${propName.toLowerCase()}`,
                        children: typeText,
                    },
                ];
            }
        }
        return typeText;
    }
    inlineType(prop) {
        let typeNode = undefined;
        const type = typeof prop.type === 'string'
            ? prop.type
            : prop.parent
                ? `:${prop.parent}`
                : undefined;
        if (typeof type === 'string') {
            typeNode = this.propLink(type);
        }
        else if (prop.kind) {
            typeNode = [
                {
                    type: 'text',
                    value: `${api.PropKind[prop.kind].toLowerCase()}`,
                },
            ];
        }
        if (prop.name) {
            return [
                {
                    type: 'text',
                    value: `${prop.name}${typeNode ? ': ' : ''}`,
                },
                ...(typeNode || []),
            ];
        }
        return typeNode || [];
    }
    typeNode(prop, showValue = false) {
        const type = typeof prop.type === 'string'
            ? prop.type
            : prop.parent
                ? `:${prop.parent}`
                : undefined;
        if (typeof type === 'string') {
            return this.propLink(type);
        }
        if (showValue && api.hasValue(prop) && prop.value !== undefined) {
            return [
                {
                    type: 'text',
                    value: prop.value.toString(),
                },
            ];
        }
        if (prop.kind) {
            return [
                {
                    type: 'text',
                    value: `${api.PropKind[prop.kind].toLowerCase()}`,
                },
            ];
        }
        if (prop.name) {
            return [
                {
                    type: 'text',
                    value: prop.name,
                },
            ];
        }
        return [];
    }
    extractPropType(prop, expandGenerics = true, showValue = false) {
        var _a, _b;
        if (api.isUnionProp(prop) && prop.properties) {
            return [
                {
                    type: 'paragraph',
                    children: (_a = prop.properties) === null || _a === void 0 ? void 0 : _a.reduce((acc, t, idx) => {
                        const r = [...acc, ...this.extractPropType(t, true, true)];
                        if (prop.properties && idx < prop.properties.length - 1) {
                            r.push({ type: 'text', value: ' | ' });
                        }
                        return r;
                    }, []),
                },
            ];
        }
        else if (api.isClassLikeProp(prop)) {
            if (prop.properties) {
                const typeArguments = expandGenerics
                    ? prop.properties.reduce((acc, p, idx) => {
                        const result = [...acc, ...this.inlineType(p)];
                        if (prop.properties && idx < prop.properties.length - 1) {
                            result.push({
                                type: 'text',
                                value: ', ',
                            });
                        }
                        return result;
                    }, [])
                    : [];
                const result = this.propLink(typeof prop.type === 'string' ? prop.type : prop.name);
                if (typeArguments.length) {
                    result.push(...[
                        {
                            type: 'text',
                            value: '<',
                        },
                        ...typeArguments,
                        {
                            type: 'text',
                            value: '>',
                        },
                    ]);
                }
                return result;
            }
            else {
                const genericArguments = (_b = prop.generics) === null || _b === void 0 ? void 0 : _b.reduce((acc, p, idx) => {
                    const result = [...acc, ...this.inlineType(p)];
                    if (prop.generics && idx < prop.generics.length - 1) {
                        result.push({
                            type: 'text',
                            value: ', ',
                        });
                    }
                    return result;
                }, []);
                const result = this.typeNode(prop);
                if (genericArguments === null || genericArguments === void 0 ? void 0 : genericArguments.length) {
                    result.push(...[
                        {
                            type: 'text',
                            value: '<',
                        },
                        ...genericArguments,
                        {
                            type: 'text',
                            value: '>',
                        },
                    ]);
                }
                return result;
            }
        }
        else if (api.isArrayProp(prop) && prop.properties) {
            return [
                {
                    type: 'paragraph',
                    children: [
                        ...prop.properties.reduce((acc, p, idx) => {
                            const result = this.extractPropType(p);
                            if (prop.properties && idx < prop.properties.length - 1) {
                                result.push({
                                    type: 'text',
                                    value: ', ',
                                });
                            }
                            return [...acc, ...result];
                        }, []),
                        { type: 'text', value: '[]' },
                    ],
                },
            ];
        }
        else if (api.isTupleProp(prop) && prop.properties) {
            return [
                {
                    type: 'paragraph',
                    children: [
                        { type: 'text', value: '[' },
                        ...prop.properties.reduce((acc, p, idx) => {
                            const result = this.extractPropType(p);
                            if (prop.properties && idx < prop.properties.length - 1) {
                                result.push({
                                    type: 'text',
                                    value: ', ',
                                });
                            }
                            return [...acc, ...result];
                        }, []),
                        { type: 'text', value: ']' },
                    ],
                },
            ];
        }
        else {
            return this.typeNode(prop, showValue);
        }
    }
    extractTSType(prop) {
        const result = [];
        result.push({
            type: 'heading',
            depth: 2,
            children: [{ type: 'text', value: prop.name }],
        });
        if (prop.description) {
            result.push(...prop.description.split('\n').map((d) => ({
                type: 'paragraph',
                children: [
                    {
                        type: 'text',
                        value: d,
                    },
                ],
            })));
        }
        if (api.isFunctionProp(prop)) {
            result.push(...this.extractFunction(prop));
        }
        else if (api.hasProperties(prop)) {
            result.push(...this.extractInterface(prop));
        }
        return result;
    }
    extract() {
        const result = [];
        if (this.files) {
            const props = api.parseFiles(this.files, {
                collectFilePath: true,
                extractNames: this.names,
                consolidateParents: true,
                plugins: [propTypesPlugin__default['default'], reactPlugin__default['default']],
            });
            Object.keys(props).forEach((key) => {
                if (key !== '__parents' && key !== '__diagnostics') {
                    this.topLevelProps[key] = props[key];
                }
            });
            const parents = props['__parents'];
            if (parents) {
                Object.keys(parents).forEach((key) => {
                    this.topLevelProps[key] = parents[key];
                });
            }
            Object.keys(this.topLevelProps).forEach((key) => {
                const prop = this.topLevelProps[key];
                const nodes = this.extractTSType(prop);
                result.push(...nodes);
            });
        }
        return result;
    }
}

/* eslint-disable prefer-spread */
const insertTSDoc = (fileName) => () => {
    const resolve = (file) => {
        if (file.startsWith('.')) {
            return path__default['default'].resolve(path__default['default'].dirname(path__default['default'].resolve(fileName)), file);
        }
        return require.resolve(file);
    };
    const splitCommaAttribute = (attributes, name) => {
        if (attributes) {
            const attribute = attributes.find((attribute) => attribute[0] === name);
            return attribute
                ? attribute[1]
                    .split(',')
                    .map((f) => f.trim())
                    .filter((f) => f)
                : undefined;
        }
        return undefined;
    };
    return (node) => {
        const sections = extractCustomTag(node, 'api-readme');
        if (sections) {
            sections.forEach(({ attrs, attributes = [] }) => {
                var _a;
                if (attributes) {
                    const newNodes = [];
                    const files = ((_a = splitCommaAttribute(attributes, 'files')) === null || _a === void 0 ? void 0 : _a.map((f) => resolve(f))) ||
                        [];
                    if (!files.length) {
                        const srcFolder = resolve('./src');
                        if (fs__default['default'].existsSync(srcFolder)) {
                            const sourceFiles = fs__default['default'].readdirSync(srcFolder);
                            const indexFile = sourceFiles.find((f) => f.startsWith('index'));
                            if (indexFile) {
                                files.push(resolve(path__default['default'].resolve(srcFolder, indexFile)));
                            }
                        }
                    }
                    if (files.length) {
                        const extractProps = new ExtractProps(files, splitCommaAttribute(attributes, 'extract'));
                        const tsNodes = extractProps.extract();
                        newNodes.push(...tsNodes);
                    }
                    inlineNewContent(attrs, newNodes);
                }
            });
        }
    };
};

/**
 * create markdown pages from README.md type files
 * command-line arguments
 * @param {string} f - input file name to be processed, default is README.md
 * @param {boolean} t - whether to generate a table of contents, default is true
 * @param {boolean} l - whether to output logs, default is true
 */
var index = () => {
    const args = yargs__namespace
        .help()
        .alias('help', 'h')
        .option('f', {
        alias: 'file',
        describe: 'Input file name',
        type: 'string',
        default: 'README.md',
    })
        .option('t', {
        alias: 'toc',
        describe: 'Table of content',
        type: 'boolean',
        default: true,
    })
        .option('l', {
        alias: 'log',
        describe: 'Enabled logs',
        type: 'boolean',
        default: true,
    });
    const options = args.parse(process.argv);
    const fileName = options.f;
    if (options.log) {
        console.log('processing file:', chalk__default['default'].red(path__default['default'].resolve(fileName)));
    }
    let r = remark__default['default']().use(insertTSDoc(fileName)).use(insertOverview);
    if (options.toc) {
        r = r.use(toc__default['default'], { tight: true });
    }
    r.process(fs__default['default'].readFileSync(fileName, 'utf8'), function (err, file) {
        if (err) {
            throw err;
        }
        fs__default['default'].writeFileSync(fileName, String(file));
    });
};

module.exports = index;
